Object subclass:#JavaObject
	instanceVariableNames:''
	classVariableNames:''
	poolDictionaries:''
	category:'Java-Classes'
!


!JavaObject class methodsFor:'misc'!

resolveClassRefs
!

resolveClassRefsIgnoring:setOfClasses
! !

!JavaObject class methodsFor:'smalltalk interface'!

convertJavaObject:val signature:retValSignature
^ val.
    retValSignature = 'void' ifTrue:[
        ^ #void
    ].
    retValSignature = 'boolean' ifTrue:[
        val == 0 ifTrue:[^ false].
        ^ true
    ].
    retValSignature = 'int' ifTrue:[
        val isInteger ifFalse:[
            self halt
        ].
        ^ val
    ].
    retValSignature = 'char[]' ifTrue:[
        "/ these are ST-strings
        ^ val
    ].

    retValSignature = 'char' ifTrue:[
        "/ these are ST-characters
        val isInteger ifTrue:[
            ^ Character value:val
        ].
        self halt.
        ^ val
    ].

    retValSignature = 'Object' ifTrue:[
        ^ val
    ].

    retValSignature = 'String' ifTrue:[
        ^ self stringFromJavaString:val
    ].

    'no conversion for: ' print. val class name print. ' to: ' print. retValSignature printNL.
    ^ val.


!

javaStringFrom:aString
    "hard-coding internas of java.lang.String here is bad ..."

    |s|

    s := Java java_lang_String basicNew.
    s instVarNamed:'value'  put: aString.
    s instVarNamed:'offset' put: 0.
    s instVarNamed:'count'  put: aString size.
    ^ s

    "Modified: 26.3.1997 / 13:43:54 / cg"
!

stringFromJavaString:aJavaString
    "hard-coding internas of java.lang.String here is bad ..."

    |start stop|

    "/ start := (aJavaString instVarNamed:'offset') + 1.
    start := (aJavaString instVarAt:(JavaSlotIndexCache string_slot_offset)) + 1.

    "/ stop := start + (aJavaString instVarNamed:'count') - 1.
    stop := start + (aJavaString instVarAt:(JavaSlotIndexCache string_slot_count)) - 1.

    "/ ^ ((aJavaString instVarNamed:'value') copyFrom:start to:stop) asString
    ^ ((aJavaString instVarAt:(JavaSlotIndexCache string_slot_value)) 
                copyFrom:start to:stop) asString

    "Modified: 8.2.1997 / 13:46:22 / cg"
! !

!JavaObject methodsFor:'initialization'!

initializeFields:initialValues
    |sz|

    sz := self class instSize.
    1 to:sz do:[:i |
        self instVarAt:i put:(initialValues at:i)
    ]
!

initializeToZero
    |sz|

    sz := self class instSize.
    1 to:sz do:[:i |
        self instVarAt:i put:0
    ]
! !

!JavaObject methodsFor:'message sending'!

doesNotUnderstand:aMessage
    |args numArgs javaMethod sel arg1 arg2|

    args := aMessage arguments.
    numArgs := args size.
    sel := aMessage selector.

    numArgs == 0 ifTrue:[
        ^ self invoke:sel.
    ].
    arg1 := args at:1.
    numArgs == 1 ifTrue:[
        ^ self invoke:sel with:arg1.
    ].
    arg2 := args at:2.
    numArgs == 2 ifTrue:[
        ^ self invoke:sel with:arg1 with:arg2.
    ].
    ^ super doesNotUnderstand:aMessage

    "Modified: 18.3.1997 / 11:47:16 / cg"
!

invokeJava:selector
    "send javaSelector (name+sig) message, without arguments"

    |method|

    method := self class lookupMethodFor:selector.
    method notNil ifTrue:[
        ^ self invokeJavaMethod:method sender:thisContext
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 1.2.1997 / 21:47:11 / cg"
!

invokeJava:selector interpreter:i sender:s with:arg1 with:arg2
    "send javaSelector (name+sig) message, with 2 arguments"

    |method|

    method := self class lookupMethodFor:selector.
    method notNil ifTrue:[
        ^ self 
            invokeJavaMethod:method 
            interpreter:i 
            sender:s 
            with:arg1 
            with:arg2
    ].
    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 18.3.1997 / 11:47:37 / cg"
!

invokeJava:selector with:arg
    "send javaSelector (name+sig) message, with 1 argument"

    |method|

    method := self class lookupMethodFor:selector.
    method notNil ifTrue:[
        ^ self 
            invokeJavaMethod:method 
            with:arg
    ].
    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 18.3.1997 / 11:47:46 / cg"
!

invokeJavaMethod:aJavaMethod
    "invoke java method, without arguments"

    ^ self 
        invokeJavaMethod:aJavaMethod 
        interpreter:JavaInterpreter new 
        sender:thisContext sender.

    "Modified: 18.3.1997 / 11:46:07 / cg"
!

invokeJavaMethod:aJavaMethod interpreter:i sender:aContext
    "invoke a java method, without arguments"

    |val|

    aJavaMethod numArgs ~~ 0 ifTrue:[
        self halt:'need arguments'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

"/    i := JavaInterpreter new.
    i push:self.

    val := i interpret:aJavaMethod sender:aContext.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)
!

invokeJavaMethod:aJavaMethod interpreter:i sender:s with:arg1
    "invoke a java method, with two arguments.
     CAVEAT: these cannot be long or doubles currently."

    |val|

    aJavaMethod numArgs ~~ 1 ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

    i push:self.
    i push:arg1.

    val := i interpret:aJavaMethod sender:s.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)

!

invokeJavaMethod:aJavaMethod interpreter:i sender:s with:arg1 with:arg2
    "invoke a java method, with two arguments.
     CAVEAT: these cannot be long or doubles currently."

    |val|

    aJavaMethod numArgs ~~ 2 ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

    i push:self.
    i push:arg1.
    i push:arg2.

    val := i interpret:aJavaMethod sender:s.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)

!

invokeJavaMethod:aJavaMethod interpreter:i sender:s with:arg1 with:arg2 with:arg3
    "invoke a java method, with 3 arguments.
     CAVEAT: these cannot be long or doubles currently."

    |val|

    aJavaMethod numArgs ~~ 3 ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

    i push:self.
    i push:arg1.
    i push:arg2.
    i push:arg3.

    val := i interpret:aJavaMethod sender:s.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)

!

invokeJavaMethod:aJavaMethod interpreter:i sender:s with:arg1 with:arg2 with:arg3 with:arg4
    "invoke a java method, with 4 arguments.
     CAVEAT: these cannot be long or doubles currently."

    |val|

    aJavaMethod numArgs ~~ 4 ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

    i push:self.
    i push:arg1.
    i push:arg2.
    i push:arg3.
    i push:arg4.

    val := i interpret:aJavaMethod sender:s.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)

!

invokeJavaMethod:aJavaMethod interpreter:i sender:s withAll:args
    "invoke a java method, with two arguments.
     CAVEAT: these cannot be long or doubles currently."

    |val|

    aJavaMethod numArgs ~~ args size ifTrue:[
        self halt:'argument count'
    ].
    aJavaMethod isStatic ifTrue:[
        self halt:'static function'
    ].

    i push:self.
    args do:[:arg |
        i push:arg.
    ].

    val := i interpret:aJavaMethod sender:s.

    ^ JavaObject convertJavaObject:val signature:(aJavaMethod retValSignature)

!

invokeJavaMethod:aJavaMethod sender:aContext
    "invoke a java method, without arguments"

    ^ self invokeJavaMethod:aJavaMethod interpreter:JavaInterpreter new sender:aContext
!

invokeJavaMethod:aJavaMethod with:arg
    "invoke a java method, with one argument.
     CAVEAT: this cannot be a long or doubles currently."

    ^ self invokeJavaMethod:aJavaMethod interpreter:JavaInterpreter new sender:thisContext sender with:arg

!

invokeJavaMethod:aJavaMethod with:arg1 with:arg2
    "invoke a java method, with two arguments.
     CAVEAT: these cannot be long or doubles currently."

    ^ self invokeJavaMethod:aJavaMethod interpreter:JavaInterpreter new sender:thisContext sender with:arg1 with:arg2

! !

!JavaObject methodsFor:'printing & storing'!

displayString
    self class == Java java_lang_String ifTrue:[
        ^ '''' , (JavaObject stringFromJavaString:self) , ''''
    ].

    ^ super displayString.

    Object errorSignal handle:[:ex |
        ^ super displayString.
    ] do:[
        ^ JavaObject stringFromJavaString:(self invoke:#toString).
    ].

    ^ super displayString. "/ ^ 'a JavaObject(' , self class name , ')'

    "Modified: 26.3.1997 / 13:44:04 / cg"
!

javaDisplayString
    Object errorSignal handle:[:ex |
        ex return.
    ] do:[
        ^ JavaObject stringFromJavaString:(self invoke:#toString).
    ].

    ^ super displayString.

    "Created: 7.4.1997 / 17:38:13 / cg"
! !

!JavaObject methodsFor:'queries'!

isJavaObject
    ^ true

    "Created: 26.3.1997 / 13:34:17 / cg"
! !

!JavaObject methodsFor:'smalltalk interface'!

invoke:selector
    "send a message, without args"

    ^ self invoke:selector interpreter:(JavaInterpreter new) sender:thisContext sender

    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#'<init>'. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#isEmpty. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#size. 
    "
!

invoke:selector interpreter:i sender:aContext
    "send a message, without args"

    |method cls sel|

    method := self lookupMethod:selector numArgs:0.
    method notNil ifTrue:[
        ^ self invokeJavaMethod:method interpreter:i sender:aContext
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#'<init>'. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#isEmpty. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#size. 
    "

    "Modified: 1.2.1997 / 21:47:25 / cg"
!

invoke:selector interpreter:i sender:aContext with:argument
    "send a message, with 1 argument. 
     TEMPORARY: This method needs more work."

    |sel method cls argClass jSel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        "/
        "/ hard to do - must find a matching method probably
        "/
        (argument isKindOf:JavaObject) ifTrue:[
            argClass := argument class.
        ] ifFalse:[
            "/
            "/ map to Java:
            "/   String -> [c
            "/
            (argument isMemberOf:String) ifTrue:[
                jSel := (selector , '([C)V') asSymbolIfInterned.
                jSel notNil ifTrue:[
                    ^ self invokeJava:jSel interpreter:i sender:aContext with:argument
                ]
            ]
        ].

        method := self lookupMethod:sel numArgs:1.
        method notNil ifTrue:[
            ^ self invokeJavaMethod:method interpreter:i sender:aContext with:argument
        ].

"/        cls := self class.
"/        [cls notNil and:[cls ~~ JavaObject]] whileTrue:[
"/            cls methodDictionary keysAndValuesDo:[:sel :aMethod |
"/                aMethod name == sel ifTrue:[
"/                    aMethod numArgs == 1 ifTrue:[
"/                        "/
"/                        "/ this is not completely correct:
"/                        "/ must look for the best type-match,
"/                        "/ (especially: have to look for best match
"/                        "/  over whole superclass chain ...)
"/                        "/ for now take the first with matching number of args
"/                        "/
"/                        ^ self invokeJavaMethod:aMethod with:argument
"/                    ]
"/                ]
"/            ].
"/            cls := cls superclass.
"/        ].
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
     stack invoke:#push with:2. 
     stack invoke:#pop. 
     stack invoke:#pop. 
     stack invoke:#pop. 
    "

    "Modified: 22.3.1997 / 00:57:33 / cg"
!

invoke:selector interpreter:i sender:aContext with:arg1 with:arg2 
    "send a message, with 2 arguments. 
     TEMPORARY: This method needs more work."

    |sel method cls argClass jSel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        method := self lookupMethod:sel numArgs:2.
        method notNil ifTrue:[
            ^ self invokeJavaMethod:method interpreter:i sender:aContext 
                               with:arg1 with:arg2
        ].
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 1.2.1997 / 21:47:31 / cg"
!

invoke:selector interpreter:i sender:aContext with:arg1 with:arg2 with:arg3
    "send a message, with 3 arguments. 
     TEMPORARY: This method needs more work."

    |sel method cls argClass jSel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        method := self lookupMethod:sel numArgs:3.
        method notNil ifTrue:[
            ^ self invokeJavaMethod:method interpreter:i sender:aContext 
                               with:arg1 with:arg2 with:arg3
        ].
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 1.2.1997 / 21:47:33 / cg"
!

invoke:selector interpreter:i sender:aContext with:arg1 with:arg2 with:arg3 with:arg4
    "send a message, with 4 arguments. 
     TEMPORARY: This method needs more work."

    |sel method cls argClass jSel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        method := self lookupMethod:sel numArgs:4.
        method notNil ifTrue:[
            ^ self invokeJavaMethod:method interpreter:i sender:aContext 
                               with:arg1 with:arg2 with:arg3 with:arg4
        ].
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 1.2.1997 / 21:47:36 / cg"
!

invoke:selector interpreter:i sender:aContext withAll:arguments
    "send a message, with arguments. 
     TEMPORARY: This method needs more work."

    |sel method cls argClass jSel|

    sel := selector asSymbolIfInterned.
    sel notNil ifTrue:[
        method := self lookupMethod:sel numArgs:arguments size.
        method notNil ifTrue:[
            ^ self invokeJavaMethod:method interpreter:i sender:aContext withAll:arguments
        ].
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "Modified: 1.2.1997 / 21:47:38 / cg"
!

invoke:selector sender:aContext
    "send a message, without args"

    |method cls sel|

    method := self lookupMethod:selector numArgs:0.
    method notNil ifTrue:[
        ^ self invokeJavaMethod:method interpreter:(JavaInterpreter new) sender:aContext
    ].

    ^ super doesNotUnderstand:(Message selector:selector)

    "
     |stack|

     stack := (Java at:'java/util/Stack') basicNew.
     stack invoke:#'<init>'. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#isEmpty. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#size. 
    "

    "Modified: 1.2.1997 / 21:47:40 / cg"
!

invoke:selector with:arg
    "send a message, with 1 argument. 
     TEMPORARY: This method needs more work."

    ^ self invoke:selector interpreter:JavaInterpreter new sender:thisContext sender with:arg

    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
     stack invoke:#push with:2. 
     stack invoke:#pop. 
     stack invoke:#pop. 
     stack invoke:#pop. 
    "
!

invoke:selector with:arg1 with:arg2
    "send a message, with 2 arguments. 
     TEMPORARY: This method needs more work."

    ^ self 
        invoke:selector 
        interpreter:JavaInterpreter new 
        sender:thisContext sender 
        with:arg1
        with:arg2

    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
    "
    "
     |stack|

     stack := (Java at:'java/util/Stack') new.
     stack invoke:#push with:1. 
     stack invoke:#push with:2. 
     stack invoke:#pop. 
     stack invoke:#pop. 
     stack invoke:#pop. 
    "
!

lookupMethod:selector numArgs:nargs
    "lookup a method"

    |method cls sel|

    sel := selector.
    (sel includes:$:) ifTrue:[
        sel := sel copyTo:(sel indexOf:$:)-1    
    ].

    sel := sel asSymbolIfInterned.
    sel notNil ifTrue:[
        cls := self class.
        [cls notNil and:[cls ~~ JavaObject]] whileTrue:[
            cls methodDictionary keysAndValuesDo:[:jSel :aMethod |
                (jSel == sel ) ifTrue:[
                    ^ aMethod
                ]
            ].
            cls methodDictionary keysAndValuesDo:[:jSel :aMethod |
                (aMethod name = sel 
                or:[aMethod signatureNameWithoutReturnType = sel]) ifTrue:[
                    aMethod numArgs == nargs ifTrue:[
                        ^ aMethod
                    ]
                ]
            ].
            cls := cls superclass.
        ].
    ].

    ^ nil

    "
     |stack|

     stack := (Java at:'java.util.Stack') basicNew.
     stack lookupMethod:#'<init>' numArgs:0. 
    "
    "
     |stack|

     stack := (Java at:'java.util.Stack') new.
     stack lookupMethod:#isEmpty numArgs:0. 
    "
    "
     |frame|

     frame := (Java at:'java.awt.Frame') new.
     frame lookupMethod:#'<init> (String)' numArgs:1. 
    "

    "Modified: 22.3.1997 / 00:56:54 / cg"
! !

!JavaObject class methodsFor:'documentation'!

version
    ^ '$Header: /home/jv/Projects/SmalltalkX/repositories/cvs/stx/libjava/JavaObject.st,v 1.27 1997/04/07 17:25:44 cg Exp $'
! !
